"""Command implementation."""
from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, TypeVar

T = TypeVar("T")


class Command(Generic[T]):
    """The object for coap commands."""

    def __init__(
        self,
        method: str,
        path: list[str],
        data: Any | None = None,
        *,
        parse_json: bool = True,
        observe: bool = False,
        observe_duration: int = 0,
        process_result: Callable[..., T] | None = None,
        err_callback: Callable[[Exception], None] | None = None,
    ) -> None:
        """Create object of class."""
        self._method = method
        self._path = path
        self._data = data
        self._parse_json = parse_json
        self._process_result = process_result
        self._err_callback = err_callback
        self._observe = observe
        self._observe_duration = observe_duration
        self._raw_result: list[Any] | dict[Any, Any] | str | None = None
        # If there's no process_result callback, the result will always be None.
        # And in that case T will also be None.
        self._result: T = None  # type: ignore[assignment]

    @property
    def method(self) -> str:
        """Return method."""
        return self._method

    @property
    def path(self) -> list[str]:
        """Return path."""
        return self._path

    @property
    def data(self) -> dict[str, Any] | None:
        """Return data."""
        return self._data

    @property
    def parse_json(self) -> bool:
        """Json parsing result."""
        return self._parse_json

    def process_result(self, result: list[Any] | dict[Any, Any] | str | None) -> T:
        """Process and set result."""
        if self._process_result:
            self._result = self._process_result(result)

        self._raw_result = result
        return self._result

    @property
    def err_callback(self) -> Callable[[Exception], None] | None:
        """Will be fired when an observe request fails."""
        return self._err_callback

    @property
    def observe(self) -> bool:
        """Observe function."""
        return self._observe

    @property
    def observe_duration(self) -> int:
        """Return duration period of observations."""
        return self._observe_duration

    @property
    def raw_result(self) -> list[Any] | dict[Any, Any] | str | None:
        """Return raw result."""
        return self._raw_result

    @property
    def result(self) -> T:
        """Return result."""
        return self._result

    @property
    def path_str(self) -> str:
        """Return coap path."""
        return "/".join(str(v) for v in self._path)

    def url(self, host: str) -> str:
        """Generate url for coap client."""
        return f"coaps://{host}:5684/{self.path_str}"

    def __repr__(self) -> str:
        """Return the representation."""
        if self.data is None:
            template = "<Command {} {}{}>"
        else:
            template = "<Command {} {}: {}>"
        return template.format(self.method, self.path, self.data or "")



"""Smart tasks set timers to turn on/off lights in various ways.

> Currently supporting wake up

SmartTask # return top level info
    TaskControl # Change top level values
    StartAction # Get top level info on start action
        StartActionItem # Get info on specific device in task
            StartActionItemController # change values for task
"""
from __future__ import annotations

from datetime import datetime as dt, time, timedelta
from typing import TYPE_CHECKING, Any, List, Optional

from pydantic import BaseModel, Field

from .command import Command
from .const import (
    ATTR_DEVICE_STATE,
    ATTR_GATEWAY_INFO,
    ATTR_ID,
    ATTR_LIGHT_DIMMER,
    ATTR_REPEAT_DAYS,
    ATTR_SMART_TASK_LIGHTS_OFF,
    ATTR_SMART_TASK_NOT_AT_HOME,
    ATTR_SMART_TASK_TRIGGER_TIME_INTERVAL,
    ATTR_SMART_TASK_TRIGGER_TIME_START_HOUR,
    ATTR_SMART_TASK_TRIGGER_TIME_START_MIN,
    ATTR_SMART_TASK_TYPE,
    ATTR_SMART_TASK_WAKE_UP,
    ATTR_START_ACTION,
    ATTR_TIME_START_TIME_MINUTE,
    ATTR_TRANSITION_TIME,
    ROOT_GATEWAY,
    ROOT_SMART_TASKS,
    ROOT_START_ACTION,
)
from .resource import ApiResource, ApiResourceResponse, BaseResponse, TypeRaw
from .util import BitChoices

if TYPE_CHECKING:
    from .gateway import Gateway, GatewayInfo


WEEKDAYS: BitChoices = BitChoices(
    (
        ("mon", "Monday"),
        ("tue", "Tuesday"),
        ("wed", "Wednesday"),
        ("thu", "Thursday"),
        ("fri", "Friday"),
        ("sat", "Saturday"),
        ("sun", "Sunday"),
    )
)


class SmartTaskMixin(BaseModel):
    """Represent common task attributes."""

    state: int = Field(alias=ATTR_DEVICE_STATE)


class StartActionResponse(BaseResponse):
    """Represent a start action response."""

    transition_time: Optional[int] = Field(alias=ATTR_TRANSITION_TIME)
    dimmer: int = Field(alias=ATTR_LIGHT_DIMMER)


class TimeIntervalResponse(BaseModel):
    """Represent a time interval response."""

    hour_start: int = Field(alias=ATTR_SMART_TASK_TRIGGER_TIME_START_HOUR)
    minute_start: int = Field(alias=ATTR_TIME_START_TIME_MINUTE)


class RootStartActionResponse(SmartTaskMixin, BaseModel):
    """Represent a smart action response."""

    root_start_action: List[StartActionResponse] = Field(alias=ROOT_START_ACTION)


class SmartTaskResponse(SmartTaskMixin, ApiResourceResponse):
    """Represent a smart task response."""

    smart_task_type: int = Field(alias=ATTR_SMART_TASK_TYPE)
    repeat_days: int = Field(alias=ATTR_REPEAT_DAYS)
    start_action: RootStartActionResponse = Field(alias=ATTR_START_ACTION)
    time_interval: List[TimeIntervalResponse] = Field(
        alias=ATTR_SMART_TASK_TRIGGER_TIME_INTERVAL
    )


class SmartTask(ApiResource):
    """Represent a smart task."""

    _model_class: type[SmartTaskResponse] = SmartTaskResponse
    raw: SmartTaskResponse

    def __init__(self, gateway: Gateway, raw: TypeRaw) -> None:
        """Initialize the class."""
        super().__init__(raw)
        self._gateway = gateway
        self.delta_time_gateway_local = timedelta(0)

    @property
    def path(self) -> list[str]:
        """Return gateway path."""
        return [ROOT_SMART_TASKS, str(self.id)]

    @property
    def state(self) -> bool:
        """Boolean representing the light state of the transition."""
        return self.raw.state == 1

    @property
    def task_type_id(self) -> int:
        """Return type of task."""
        return self.raw.smart_task_type

    @property
    def task_type_name(self) -> str | None:
        """Return the task type in plain text.

        (Own interpretation of names.)
        """
        if self.is_wake_up:
            return "Wake Up"
        if self.is_not_at_home:
            return "Not At Home"
        if self.is_lights_off:
            return "Lights Off"
        return None

    @property
    def is_wake_up(self) -> bool:
        """Boolean representing if this is a wake up task."""
        return self.raw.smart_task_type == ATTR_SMART_TASK_WAKE_UP

    @property
    def is_not_at_home(self) -> bool:
        """Boolean representing if this is a not home task."""
        return self.raw.smart_task_type == ATTR_SMART_TASK_NOT_AT_HOME

    @property
    def is_lights_off(self) -> bool:
        """Boolean representing if this is a lights off task."""
        return self.raw.smart_task_type == ATTR_SMART_TASK_LIGHTS_OFF

    @property
    def repeat_days(self) -> int:
        """Return int (bit) for enabled weekdays."""
        return self.raw.repeat_days

    @property
    def repeat_days_list(self) -> list[str]:
        """Binary representation of weekdays the event takes place."""
        return WEEKDAYS.get_selected_values(self.repeat_days)

    @property
    def task_start_parameters(self) -> TimeIntervalResponse:
        """Return hour and minute that task starts."""
        return self.raw.time_interval[0]

    @property
    def task_start_time(self) -> time:
        """Return the time the task starts.

        Time is set according to iso8601.
        """
        return time(
            self.task_start_parameters.hour_start,
            self.task_start_parameters.minute_start,
        )

    @property
    def task_control(self) -> TaskControl:
        """Control a task."""
        return TaskControl(self, self.state, self.path, self._gateway)

    @property
    def start_action(self) -> StartAction:
        """Return start action object."""
        return StartAction(self, self.path)

    def __repr__(self) -> str:
        """Return a readable name for smart task."""
        state = "on" if self.state else "off"
        return f"<Task {self.id} - {self.task_type_name} - {state}>"


class TaskControl:
    """Class to control the tasks."""

    def __init__(
        self, task: SmartTask, state: bool, path: list[str], gateway: Gateway
    ) -> None:
        """Initialize TaskControl."""
        self._task = task
        self.state = state
        self.path = path
        self._gateway = gateway

    @property
    def tasks(self) -> list[StartActionItem]:
        """Return task objects of the task control."""
        return [
            StartActionItem(self._task, idx, self.state, self.path, self.raw)
            for idx in range(len(self.raw.root_start_action))
        ]

    def calibrate_time(self) -> Command[None]:
        """Calibrate difference between local time and gateway time."""

        def process_result(result: TypeRaw) -> None:
            gateway_info: GatewayInfo = GatewayInfo(result)
            if not gateway_info.current_time:
                return

            d_now = gateway_info.current_time
            d_utcnow = dt.utcnow()
            diff = d_now - d_utcnow

            self._task.delta_time_gateway_local = diff

        return Command(
            "get", [ROOT_GATEWAY, ATTR_GATEWAY_INFO], process_result=process_result
        )

    def set_dimmer_start_time(self, hour: int, minute: int) -> Command[None]:
        """Set start time for task (hh:mm) in iso8601.

        NB: dimmer starts 30 mins before time in app
        """
        new_time: dt = (
            dt(100, 1, 1, hour, minute, 00) - self._task.delta_time_gateway_local
        )

        command: dict[str, list[dict[str, int]]] = {
            ATTR_SMART_TASK_TRIGGER_TIME_INTERVAL: [
                {
                    ATTR_SMART_TASK_TRIGGER_TIME_START_HOUR: new_time.hour,
                    ATTR_SMART_TASK_TRIGGER_TIME_START_MIN: new_time.minute,
                }
            ]
        }
        return self._task.set_values(command)

    @property
    def raw(self) -> RootStartActionResponse:
        """Return raw data that it represents."""
        return self._task.raw.start_action


class StartAction:
    """Class to control the start action-node."""

    def __init__(self, smart_task: SmartTask, path: list[str]) -> None:
        """Initialize StartAction class."""
        self.smart_task = smart_task
        self.path = path

    @property
    def state(self) -> bool:
        """Return state of start action task."""
        return self.raw.state == 1

    @property
    def raw(self) -> RootStartActionResponse:
        """Return raw data that it represents."""
        return self.smart_task.raw.start_action


class StartActionItem:
    """Class to show settings for a task."""

    def __init__(
        self,
        task: SmartTask,
        index: int,
        state: bool,
        path: list[str],
        raw: RootStartActionResponse,
    ):
        """Initialize TaskInfo."""
        self.task = task
        self.index = index
        self.state = state
        self.path = path
        self._raw = raw

    @property
    def devices_list(self) -> list[dict[str, int]]:
        """Store task data for all tasks but the one we want to update."""
        output_list: list[dict[str, int]] = []
        current_data_list: list[StartActionResponse] = self._raw.root_start_action
        for idx, record in enumerate(current_data_list):
            if idx != self.index:
                list_record: dict[str, int] = {}
                list_record[ATTR_ID] = record.id
                list_record[ATTR_LIGHT_DIMMER] = record.dimmer

                if record.transition_time is not None:
                    list_record[ATTR_TRANSITION_TIME] = record.transition_time

                output_list.append(list_record)

        return output_list

    @property
    def id(self) -> int:
        """Return ID (device id) of task."""
        return self.raw.id

    @property
    def item_controller(self) -> StartActionItemController:
        """Control a task."""
        return StartActionItemController(
            self, self.raw, self.state, self.path, self.devices_list
        )

    @property
    def transition_time(self) -> int | None:
        """Transition runs for this long from the time in task_start.

        Value is in seconds x 10. Default to 0 if transition is missing.
        """
        if self.raw.transition_time is not None:
            return round(self.raw.transition_time / 60 / 10)

        return None

    @property
    def dimmer(self) -> int:
        """Return dimmer level."""
        return self.raw.dimmer

    @property
    def raw(self) -> StartActionResponse:
        """Return raw data that it represents."""
        return self._raw.root_start_action[self.index]

    def __repr__(self) -> str:
        """Return a readable name for this class."""
        return f"<StartActionItem (Device: {self.id} - Dimmer: {self.dimmer} - Time: {self.transition_time})>"


class StartActionItemController:
    """Class to edit settings for a task."""

    def __init__(
        self,
        item: StartActionItem,
        raw: StartActionResponse,
        state: bool,
        path: list[str],
        devices_list: list[dict[str, int]],
    ):
        """Initialize StartActionItemController."""
        self._item = item
        self.raw = raw
        self.state = state
        self.path = path
        self.devices_list = devices_list

    def set_dimmer(self, dimmer: int) -> Command[None]:
        """Set final dimmer value for task."""
        root_start_action_list: list[dict[str, int]] = [
            {
                ATTR_ID: self.raw.id,
                ATTR_LIGHT_DIMMER: dimmer,
            }
        ]

        if self.raw.transition_time is not None:
            root_start_action_list[0][ATTR_TRANSITION_TIME] = self.raw.transition_time

        root_start_action_list.extend(self.devices_list)

        command: dict[str, dict[str, Any]] = {
            ATTR_START_ACTION: {
                ATTR_DEVICE_STATE: int(self.state),
                ROOT_START_ACTION: root_start_action_list,
            }
        }
        return self.set_values(command)

    def set_transition_time(self, transition_time: int) -> Command[None]:
        """Set time (mins) for light transition."""
        root_start_action_list: list[dict[str, int]] = [
            {
                ATTR_ID: self.raw.id,
                ATTR_LIGHT_DIMMER: self.raw.dimmer,
                ATTR_TRANSITION_TIME: transition_time * 10 * 60,
            }
        ]
        root_start_action_list.extend(self.devices_list)

        command: dict[str, dict[str, Any]] = {
            ATTR_START_ACTION: {
                ATTR_DEVICE_STATE: int(self.state),
                ROOT_START_ACTION: root_start_action_list,
            }
        }
        return self.set_values(command)

    def set_values(self, values: dict[str, dict[str, Any]]) -> Command[None]:
        """
        Set values on task control.

        Returns a Command.
        """
        return Command("put", self._item.path, values)



"""Represent an air purifier."""
from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import Field

from ..const import (
    ATTR_AIR_PURIFIER_AIR_QUALITY,
    ATTR_AIR_PURIFIER_CONTROLS_LOCKED,
    ATTR_AIR_PURIFIER_FAN_SPEED,
    ATTR_AIR_PURIFIER_FILTER_LIFETIME_REMAINING,
    ATTR_AIR_PURIFIER_FILTER_LIFETIME_TOTAL,
    ATTR_AIR_PURIFIER_FILTER_RUNTIME,
    ATTR_AIR_PURIFIER_FILTER_STATUS,
    ATTR_AIR_PURIFIER_LEDS_OFF,
    ATTR_AIR_PURIFIER_MODE,
    ATTR_AIR_PURIFIER_MODE_AUTO,
    ATTR_AIR_PURIFIER_MOTOR_RUNTIME_TOTAL,
)
from ..resource import BaseResponse


class AirPurifierResponse(BaseResponse):
    """Represent API response for a blind."""

    air_quality: int = Field(alias=ATTR_AIR_PURIFIER_AIR_QUALITY)
    controls_locked: int = Field(alias=ATTR_AIR_PURIFIER_CONTROLS_LOCKED)
    fan_speed: int = Field(alias=ATTR_AIR_PURIFIER_FAN_SPEED)
    filter_lifetime_remaining: int = Field(
        alias=ATTR_AIR_PURIFIER_FILTER_LIFETIME_REMAINING
    )
    filter_lifetime_total: int = Field(alias=ATTR_AIR_PURIFIER_FILTER_LIFETIME_TOTAL)
    filter_runtime: int = Field(alias=ATTR_AIR_PURIFIER_FILTER_RUNTIME)
    filter_status: int = Field(alias=ATTR_AIR_PURIFIER_FILTER_STATUS)
    leds_off: int = Field(alias=ATTR_AIR_PURIFIER_LEDS_OFF)
    mode: int = Field(alias=ATTR_AIR_PURIFIER_MODE)
    motor_runtime_total: int = Field(alias=ATTR_AIR_PURIFIER_MOTOR_RUNTIME_TOTAL)


if TYPE_CHECKING:
    # avoid cyclic import at runtime.
    from . import Device


class AirPurifier:
    """Represent an air purifier."""

    def __init__(self, device: Device, index: int) -> None:
        """Create object of class."""
        self.device = device
        self.index = index

    @property
    def air_quality(self) -> int:
        """Get the current air quality measured by the air purifier.

        0..35: Good
        36..85: OK
        86..: Not good
        65535: If the fan is off or during measuring time after turning on
        """
        return self.raw.air_quality

    @property
    def controls_locked(self) -> bool:
        """Return True if physical controls on the air purifier are locked."""
        return self.raw.controls_locked == 1

    @property
    def fan_speed(self) -> int:
        """Get the current fan speed of the air purifier.

        0: Device is off
        2..50: Fan speed with a step size of 1.
        """
        return self.raw.fan_speed

    @property
    def filter_lifetime_remaining(self) -> int:
        """Return remaining lifetime of filter, expressed in minutes."""
        return self.raw.filter_lifetime_remaining

    @property
    def filter_lifetime_total(self) -> int:
        """Return total lifetime of filter, expressed in minutes."""
        return self.raw.filter_lifetime_total

    @property
    def filter_runtime(self) -> int:
        """Return filter runtime, expressed in minutes."""
        return self.raw.filter_runtime

    @property
    def filter_status(self) -> bool:
        """
        Return True if filter needs to be replaced.

        This property is true when filter_lifetime_remaining is less than zero.
        """
        return bool(self.raw.filter_status)

    @property
    def is_auto_mode(self) -> bool:
        """
        Return auto mode on or off.

        Auto mode sets the fan speed automatically based on the air quality.
        """
        return self.raw.mode == ATTR_AIR_PURIFIER_MODE_AUTO

    @property
    def leds_off(self) -> bool:
        """Return True if led's on the air purifier are turned off."""
        return self.raw.leds_off == 1

    @property
    def motor_runtime_total(self) -> int:
        """Return runtime of fan motor, expressed in minutes."""
        return self.raw.motor_runtime_total

    @property
    def raw(self) -> AirPurifierResponse:
        """Return raw data that it represents."""
        air_purifier_control_response = self.device.raw.air_purifier_control
        assert air_purifier_control_response is not None
        return air_purifier_control_response[self.index]

    @property
    def state(self) -> bool:
        """Return device state, ie on or off."""
        return self.raw.mode > 0



"""Represent a mood on the gateway."""
from __future__ import annotations

from .const import ROOT_MOODS
from .resource import ApiResource, TypeRaw


class Mood(ApiResource):
    """Mood."""

    def __init__(self, raw: TypeRaw, parent: int) -> None:
        """Create object of class."""
        super().__init__(raw)
        self._parent = parent

    @property
    def path(self) -> list[str]:
        """Path."""
        return [ROOT_MOODS, str(self._parent), str(self.id)]

    def __repr__(self) -> str:
        """Return representation of class object."""
        return f"<Mood {self._parent} {self.name}>"



"""Represent a socket."""
from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import Field

from ..const import ATTR_DEVICE_STATE
from ..resource import BaseResponse

if TYPE_CHECKING:
    # avoid cyclic import at runtime.
    from . import Device


class SocketResponse(BaseResponse):
    """Represent API response for a blind."""

    state: int = Field(alias=ATTR_DEVICE_STATE)


class Socket:
    """Represent a socket."""

    def __init__(self, device: Device, index: int) -> None:
        """Create object of class."""
        self.device = device
        self.index = index

    @property
    def state(self) -> bool:
        """State."""
        return self.raw.state == 1

    @property
    def raw(self) -> SocketResponse:
        """Return raw data that it represents."""
        socket_control_response = self.device.raw.socket_control
        assert socket_control_response is not None
        return socket_control_response[self.index]

    def __repr__(self) -> str:
        """Return representation of class object."""
        state = "on" if self.state else "off"
        return f"<Socket #{self.index} - name: {self.device.name}, state: {state}>"



"""Represent a blind."""
from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import Field

from ..const import ATTR_BLIND_CURRENT_POSITION
from ..resource import BaseResponse


class BlindResponse(BaseResponse):
    """Represent API response for a blind."""

    current_cover_position: int = Field(alias=ATTR_BLIND_CURRENT_POSITION)


if TYPE_CHECKING:
    # avoid cyclic import at runtime.
    from . import Device


class Blind:
    """Represent a blind."""

    def __init__(self, device: Device, index: int) -> None:
        """Create object of class."""
        self.device = device
        self.index = index

    @property
    def raw(self) -> BlindResponse:
        """Return raw data that it represents."""
        blind_control_response = self.device.raw.blind_control
        assert blind_control_response is not None
        return blind_control_response[self.index]

    @property
    def current_cover_position(self) -> int:
        """Get the current position of the blind."""
        return self.raw.current_cover_position



"""Classes to interact with devices."""
from __future__ import annotations

from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel, Field

from ..const import (
    ATTR_APPLICATION_TYPE,
    ATTR_DEVICE_BATTERY,
    ATTR_DEVICE_FIRMWARE_VERSION,
    ATTR_DEVICE_INFO,
    ATTR_DEVICE_MANUFACTURER,
    ATTR_DEVICE_MODEL_NUMBER,
    ATTR_DEVICE_POWER_SOURCE,
    ATTR_DEVICE_SERIAL,
    ATTR_LAST_SEEN,
    ATTR_LIGHT_CONTROL,
    ATTR_REACHABLE_STATE,
    ATTR_START_BLINDS,
    ATTR_SWITCH_PLUG,
    ROOT_AIR_PURIFIER,
    ROOT_DEVICES,
    ROOT_SIGNAL_REPEATER,
)
from ..device.light import LightResponse
from ..resource import ApiResource, ApiResourceResponse
from .air_purifier import AirPurifierResponse
from .air_purifier_control import AirPurifierControl
from .blind import BlindResponse
from .blind_control import BlindControl
from .light_control import LightControl
from .signal_repeater import SignalRepeaterResponse
from .signal_repeater_control import SignalRepeaterControl
from .socket import SocketResponse
from .socket_control import SocketControl


class DeviceInfoResponse(BaseModel):
    """Represent the device info part of the device response."""

    manufacturer: str = Field(alias=ATTR_DEVICE_MANUFACTURER)
    model_number: str = Field(alias=ATTR_DEVICE_MODEL_NUMBER)
    serial: str = Field(alias=ATTR_DEVICE_SERIAL)
    firmware_version: str = Field(alias=ATTR_DEVICE_FIRMWARE_VERSION)
    power_source: Optional[int] = Field(alias=ATTR_DEVICE_POWER_SOURCE)
    battery_level: Optional[int] = Field(alias=ATTR_DEVICE_BATTERY)


class DeviceResponse(ApiResourceResponse):
    """Represent a device response."""

    air_purifier_control: Optional[List[AirPurifierResponse]] = Field(
        alias=ROOT_AIR_PURIFIER
    )
    application_type: int = Field(alias=ATTR_APPLICATION_TYPE)
    blind_control: Optional[List[BlindResponse]] = Field(alias=ATTR_START_BLINDS)
    device_info: DeviceInfoResponse = Field(alias=ATTR_DEVICE_INFO)
    last_seen: Optional[int] = Field(alias=ATTR_LAST_SEEN)
    light_control: Optional[List[LightResponse]] = Field(alias=ATTR_LIGHT_CONTROL)
    reachable: int = Field(alias=ATTR_REACHABLE_STATE)
    signal_repeater_control: Optional[List[SignalRepeaterResponse]] = Field(
        alias=ROOT_SIGNAL_REPEATER
    )
    socket_control: Optional[List[SocketResponse]] = Field(alias=ATTR_SWITCH_PLUG)


class Device(ApiResource):
    """Base class for devices."""

    _model_class: type[DeviceResponse] = DeviceResponse
    raw: DeviceResponse

    @property
    def application_type(self) -> int:
        """Return application type."""
        return self.raw.application_type

    @property
    def path(self) -> list[str]:
        """Return path."""
        return [ROOT_DEVICES, str(self.id)]

    @property
    def device_info(self) -> "DeviceInfo":
        """Return Device information."""
        return DeviceInfo(self)

    @property
    def last_seen(self) -> datetime | None:
        """Return timestamp when last seen."""
        last_seen = self.raw.last_seen

        if last_seen is not None:
            return datetime.utcfromtimestamp(last_seen)

        return None

    @property
    def reachable(self) -> bool:
        """Check if gateway is reachable."""
        return self.raw.reachable == 1

    @property
    def has_light_control(self) -> bool:
        """Check if light_control is present."""
        return self.raw.light_control is not None

    @property
    def light_control(self) -> LightControl | None:
        """Return light_control."""
        if self.has_light_control:
            return LightControl(self)
        return None

    @property
    def has_socket_control(self) -> bool:
        """Check if socket_control is present."""
        return self.raw.socket_control is not None

    @property
    def socket_control(self) -> SocketControl | None:
        """Return socket_control."""
        if self.has_socket_control:
            return SocketControl(self)
        return None

    @property
    def has_blind_control(self) -> bool:
        """Check if blind_control is present."""
        return self.raw.blind_control is not None

    @property
    def blind_control(self) -> BlindControl | None:
        """Return blind_control."""
        if self.has_blind_control:
            return BlindControl(self)
        return None

    @property
    def has_signal_repeater_control(self) -> bool:
        """Check if signal_repeater_control is present."""
        return self.raw.signal_repeater_control is not None

    @property
    def signal_repeater_control(self) -> SignalRepeaterControl | None:
        """Return signal_repeater control, if any."""
        if self.has_signal_repeater_control:
            return SignalRepeaterControl(self)
        return None

    @property
    def has_air_purifier_control(self) -> bool:
        """Check if air_purifier_control is present."""
        return self.raw.air_purifier_control is not None

    @property
    def air_purifier_control(self) -> AirPurifierControl | None:
        """Return air_purifier control, if any."""
        if self.has_air_purifier_control:
            return AirPurifierControl(self)
        return None

    def __repr__(self) -> str:
        """Return representation of class object."""
        return f"<{self.id} - {self.name} ({self.device_info.model_number})>"


class DeviceInfo:
    """Represent device information.

    http://www.openmobilealliance.org/tech/profiles/LWM2M_Device-v1_0.xml
    """

    VALUE_POWER_SOURCES = {
        1: "Internal Battery",
        2: "External Battery",
        3: "Battery",  # Not in spec, used by remote
        4: "Power over Ethernet",
        5: "USB",
        6: "AC (Mains) power",
        7: "Solar",
    }

    def __init__(self, device: Device) -> None:
        """Create object of class."""
        self._device = device

    @property
    def manufacturer(self) -> str:
        """Human readable manufacturer name."""
        return self.raw.manufacturer

    @property
    def model_number(self) -> str:
        """Return model identifier string (manufacturer specified string)."""
        return self.raw.model_number

    @property
    def serial(self) -> str:
        """Return serial string."""
        return self.raw.serial

    @property
    def firmware_version(self) -> str:
        """Return current firmware version of device."""
        return self.raw.firmware_version

    @property
    def power_source(self) -> int | None:
        """Power source."""
        return self.raw.power_source

    @property
    def power_source_str(self) -> str | None:
        """Represent current power source."""
        if self.raw.power_source is not None:
            return self.VALUE_POWER_SOURCES.get(self.raw.power_source, "Unknown")

        return None

    @property
    def battery_level(self) -> int | None:
        """Battery in 0..100."""
        return self.raw.battery_level

    @property
    def raw(self) -> DeviceInfoResponse:
        """Return raw data that it represents."""
        return self._device.raw.device_info



"""Represent a signal repeater."""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..resource import BaseResponse

if TYPE_CHECKING:
    from . import Device


class SignalRepeaterResponse(BaseResponse):
    """Represent API response for a signal repeater."""


class SignalRepeater:
    """Represent a signal repeater."""

    def __init__(self, device: Device, index: int) -> None:
        """Create object of class."""
        self.device = device
        self.index = index

    @property
    def id(self) -> int:
        """Return ID."""
        return self.raw.id

    @property
    def raw(self) -> SignalRepeaterResponse:
        """Return raw data that it represents."""
        signal_repeater_control_response = self.device.raw.signal_repeater_control
        assert signal_repeater_control_response is not None
        return signal_repeater_control_response[self.index]
